var ohm = require("./ohm.js");
var set = require("./set.js");

var fs = require('fs');

var path = require('path');
var here = path.dirname(process.argv[1]);

var ptsOhm = fs.readFileSync(here + '/grammar.ohm').toString();
var grammar = ohm.makeGrammar(ptsOhm, "");

exports.grammar = grammar;
var ast = grammar.attribute({
    _default: function () {
      var l = [this.ctorName];
      for(var i in this.args) {
        l.push(ast(this.args[i])); 
      };
      return l;
    },

    ident:    function(_,_) { 
      return ['ident', this.interval.contents]; 
    }
});
exports.ast = ast;

var haskell = grammar.attribute({
  Pgm : function(stmts, expr) {
    return "(Pgm " + haskell(stmts) + " " + haskell(expr) + ")";
  },

  Stmts : function(stmts) {
    var s = "(Stmts [";
    var first = true;
    for(i in stmts.values) {
      if(!first) s += ',';
      first = false;
      s += haskell(stmts.values[i]);
    }
    return s + "])";
  },

  Stmt_Load : function(_, fname, _) {
    var base = fname.interval.contents.slice(1,-1);
    var str = fs.readFileSync("lib/" + base + ".pts", {encoding:'UTF-8'});
    return haskell(grammar.matchContents(str,'Stmts',true));
  },

  Stmt_Decl : function(_, nm, _, ty, _, bnd, _) {
    return '(SDecl "' + haskell(nm) + '" ' + haskell(ty) + ' ' + haskell(bnd) + ')';
  },
 
  Stmt_Comment : function(_,_,_,s) { return haskell(s); },

  Expr_Abs : function(_,nm,_,ty,_,body) { 
    return '(Lam "' + haskell(nm) + '" ' + haskell(ty) + ' ' + haskell(body) + ')';
  },
  Expr_Pi  : function(_,nm,_,ty,_,body) { 
    return '(Pi "' + haskell(nm) + '" ' + haskell(ty) + ' ' + haskell(body) + ')';
  },
  Expr_Let : function(_,nm,_,ty,_,bnd,_,bdy) {
    return '(App (Lam "' + haskell(nm) + '" ' + haskell(ty) + ' ' + haskell(bdy) + ') ' + haskell(bnd) + ')';
  },

  // deprecated!
  Expr_Decl : function(_,nm,_,ty,_,bnd,_,bdy) {
    return '(Decl "' + haskell(nm) + '" ' + haskell(ty) + ' ' + haskell(bnd) + ' ' + haskell(bdy) + ')';
  },

  Expr_Comment: function(_,_,_,e) { return haskell(e); },

  Expr_Arr : function(e) { return haskell(e); },
  
  ArrExpr_Arr : function(l,_,r) { return '(Pi "" ' + haskell(l) + ' ' + haskell(r) + ')'; },
  ArrExpr_Atom : function(e) { return haskell(e); },
  
  AppExpr_App : function(l,r) { return '(App ' + haskell(l) + ' ' + haskell(r) + ')'; },
  AppExpr_Atom : function(e) { return haskell(e); },

  AtomExpr_Paren : function(_,e,_) { return haskell(e); },
  AtomExpr_Quote : function(_,e,_) { return '(Quote ' + haskell(e) + ')'; },
  AtomExpr_TypedQuote : function(_,e,_) { return '(TypedQuote ' + haskell(e) + ')'; },
  AtomExpr_Var : function(nm) { return '(Var "' + haskell(nm) + '")'; },
  AtomExpr_Star : function(_) { return 'Star'; },
  AtomExpr_Box  : function(_) { return 'Box'; },

  ident : function(_,_) { return this.interval.contents; }
});

exports.haskell = haskell;


function sort(rules) {
  var sortEnv = {};
  function withSort(nm,s,f) {
    var tmp = sortEnv[nm];
    sortEnv[nm]=s;
    var ret = f();
    sortEnv[nm]=tmp;
    return ret;
    
  }
  var attr = grammar.attribute({
    lookupSort: function(ident) {
      var nm = ident.interval.contents;
      for(var i = length(sortEnv) - 1; i>=0; i--) {
        if(sortEnv[i][nm])
          return sortEnv[i][nm]
      }
    },

    Expr_Abs: function(_,ident,_,ty,_,body) {
      var nm = ident.interval.contents;
      var s1 = attr(ty);
      
      return withSort(nm,s1,function() {
        var s2 = attr(body);
        return rules[s1][s2];
      });
    },

    ArrExpr_Arr: function(l,_,r) {
      var s1 = attr(l);
      var s2 = attr(r);
      return rules[s1][s2];
    },

    AtomExpr_Var: function(ident) {
      var nm = ident.interval.contents;
      return sortEnv[nm];
    },
 
    AtomExpr_Star: function(_) { return '#'; },
    AtomExpr_Box: function(_) { return '^'; },
  });
  return attr;
}

exports.sort = function() {
  var rules = {'*' : {'*' : '*'},
               '#' : {'#' : '#', '*' : '*'},
               '^' : {'#' : '#', '*' : '*'}}
  return sort(rules);
}

// convenience functions
exports.parse = function(str,rule,action) {
  var r = grammar.matchContents(str, rule, true);
  if(r) {
    return action(r);
  }
};

exports.parseAst = function(str,rule) {
  return ast(grammar.matchContents(str, rule, true));
}

exports.parseHs = function(str,rule) {
  return haskell(grammar.matchContents(str,rule,true));
}
