module Language.SystemU.QuoteTerm where

import Language.SystemU.Syntax
import Language.SystemU.Sorted (Sort(..))
import qualified Language.SystemU.QuoteType as QuoteType
import Language.SystemU.TypeCheck hiding (app)
import qualified Language.SystemU.TypeCheck as T
import qualified Language.SystemU.Builtins as B
import qualified Language.SystemU.Coerce as C

import Control.Applicative
import Prelude hiding (pi)
import System.IO.Unsafe

type Type = Tm ()

expandTermQuotes :: Typed -> Either String (Tm ())
expandTermQuotes = expandTermQuotes' []
  
expandTermQuotes' :: Env -> Typed -> Either String (Tm ())
expandTermQuotes' env e = 
  let go = expandTermQuotes' env
      goWith nm ty = expandTermQuotes' ((nm,ty):env)
  in    
  case e of
    Lam _ v tv b   -> lam v (erase tv) <$> goWith v (erase tv) b
    App _ a b      -> app <$> go a <*> go b
    Quote _ b      -> quoteTm env b
    _              -> return $ erase e

-- Produce an (Exp t)
quoteTm :: Env -> Typed -> Either String (Tm ())
quoteTm env e = do
  let v = unsafePerformIO genSym
  e' <- quoteTm' (var v) env e
  return $ Lam () v (B.vU ==> star) e'

fromJust (Just x) = x  
  
quoteType' env t 
  | s == Type = QuoteType.quoteType st
  | otherwise = error $ unlines [
                             "quoteType' " ++ show s
                           , show t  
                           , show env
                           ]  
  where st = case typedSortInf env t of
               Just st -> st
               Nothing -> error $ unlines [
                            "quoteType': typedSortInf failed"
                          , "t = " ++ show t
                          , "env = " ++ show env
                          ]  
        s  = tmType st

quoteTm' :: Tm () -> Env -> Typed -> Either String (Tm ())
quoteTm' v env e = 
  let go = quoteTm' v env
      goWith nm t = quoteTm' v ((nm,t):env)
      quoteType = quoteType' env
      quoteTypeWith x t = quoteType' ((x,t):env)
  in    
  case e of
    Lam lt nm t b 
      | isStar s   -> do let Pi _ _ t1 t2 = lt
                             q1 = quoteType t1
                             q2 = quoteType t2
                         b' <- go b
                         return $ 
                           app4 B.vMkAbsStar v q1 q2 $
                             lam nm (app v q1) b'
      | isBox s    -> do let Pi _ _ k t = lt
                             q = quoteTypeWith nm k t
                         b' <- goWith nm k b    
                         return $ app4 B.vMkAbsBox v k 
                           (lam nm k q) (lam nm k b')
      | isDelta s  -> do let Pi _ _ _ t = lt
                             q = quoteTypeWith nm box t
                         b' <- goWith nm box b    
                         return $ app3 B.vMkAbsDelta v
                           (lam nm box q) (lam nm box b')

      where s = tmType t
    App _ a b     
      | isStar s   -> do a' <- go a
                         b' <- go b
                         let Pi _ _ t1 t2 = tmType a
                             q1 = quoteType t1
                             q2 = quoteType t2
                         return $ app5 B.vMkAppStar v q1 q2 a' b'
      | isBox s    -> do a' <- go a
                         let Pi _ nm k t = tmType a
                             q = quoteTypeWith nm k t
                             e' = app5 B.vMkAppBox v k (lam nm k q) a' (erase b)
                             c = unsafePerformIO $ 
                                   C.buildCoercion env v (tmType a) (erase b)
                         return $ app c e'    
      | isDelta s  -> do a' <- go a
                         let Pi _ nm _ t = tmType a
                             q = quoteTypeWith nm box t
                         return $ app4 B.vMkAppDelta v (lam nm box q) a' (erase b)    
      where Right s = tmType <$> typecheck env (tmType b)
    Var t x     
      | isStar s   -> return $ app3 B.vMkRep v (quoteType t) (var x)
      | otherwise  -> return $ var x                      
      where Right s = tmType <$> typecheck env t                
