module Language.SystemU.QuoteType where

-- This module expands type quotations.

import Language.SystemU.Syntax
import Language.SystemU.Sorted
import Language.SystemU.Builtins

import System.IO.Unsafe

expandTypeQuotes :: Sorted -> Tm ()
expandTypeQuotes e = 
  let go = expandTypeQuotes in
  case e of
    Var _ x        -> Var () x
    App _ f a      -> App () (go f) (go a)
    Lam _ x t b    -> Lam () x (go t) (go b)
    Pi _ x t b     -> Pi () x (go t) (go b)
    Decl _ x t d b -> Decl () x (go t) (go d) (go b)
    Quote Type t   -> norm $ quoteType $ norm t
    Quote Term t   -> Quote () (go t)
    Star _         -> Star ()
    Box _          -> Box ()
    Delta _        -> Delta ()
    _              -> error $ "expandTypeQuotes: " ++ show e    

-- Must be in normal form, and of kind *
quoteType :: Sorted -> Tm ()
quoteType ty = 
  let go = quoteType
  in  
  case ty of
    (Pi Type x l r) 
      | sort l == Type    -> app2 _ProdStar (go l) (go r)
      | sort l == Kind    -> let l' = erase l in
                             app2 _ProdBox l' $ lam x l' (go r)
      | sort l == BoxSort -> app _ProdDelta (lam x box $ go r)
    _ | sort ty == Type   -> app _Var (erase ty)

