{-# LANGUAGE TypeSynonymInstances, FlexibleInstances #-}
module Language.SystemU.Sorted (
    Sort(..)
  , Sorted  
  , Env  
  , sortInf  
  , sort
  , classifiedBy  
  ) where
       
--import qualified Language.SystemU.Core as C
import Language.SystemU.Syntax hiding (lam, app, pi)
import Control.Applicative hiding (Const)
import Prelude hiding (pi, product)

data Sort = Term | Type | Kind | BoxSort | DeltaSort
  deriving (Eq, Show)
           
type Sorted = Tm Sort

instance Eq Sorted where
  t1 == t2 = erase t1 == erase t2

{--           
sort :: Tm -> Sort
sort Star             = Kind
sort Box              = BoxSort
sort Delta            = DeltaSort
sort (Lam s _ _ _)    = s
sort (Pi s _ _ _)     = s
sort (App s _ _)      = s
sort (Var s _)        = s
sort (Decl s _ _ _ _) = s
sort (Quote s _)      = s
--}

sort :: Sorted -> Sort
sort = tmVal

product Type    Type  = Type   -- (*,*)
product Kind    Type  = Type   -- (#,*)
product BoxSort Type  = Type   -- (^,*)
product Kind    Kind  = Kind   -- (#,#)
product BoxSort Kind  = Kind   -- (^,#)
product a     b     = error $ "product " ++ show (a,b)

function Term Term  = Term     -- (*,*)
function Type Term  = Term     -- (#,*)
function Kind Term  = Term     -- (^,*)
function Type Type  = Type     -- (#,#)
function Kind Type  = Type     -- (^,#)
function a    b     = error $ "function " ++ show (a,b)

-- return the sort classfied by another sort
classifiedBy Type      = Term
classifiedBy Kind      = Type
classifiedBy BoxSort   = Kind
classifiedBy DeltaSort = BoxSort

lam :: String -> Sorted -> Sorted -> Maybe Sorted
lam nm ty body = do
  let s = function (classifiedBy $ sort ty) (sort body)
  return (Lam s nm ty body)

pi :: String -> Sorted -> Sorted -> Maybe Sorted
pi nm ty body = do
  let s = product (sort ty) (sort body)
  return (Pi s nm ty body)

app :: Sorted -> Sorted -> Maybe Sorted
app f a = do
  let s = function (sort a) (sort f)
  return (App s f a)
  
type Env = [(String,Sort)]  
  
sortInf :: Show a => Env -> Tm a -> Maybe Sorted
sortInf env t =
  let go = sortInf env
      goWith x s = sortInf ((x,s):env)
      lookup' x env = maybe (error $ unlines [ 
                                "sortInf: variable not found"
                              , x  
                              , show env  
                              ]) id $
                      lookup x env
  in    
  case t of
    Star _         -> Just $ Star Kind
    Box _          -> Just $ Box BoxSort
    Delta _        -> Just $ Delta DeltaSort
    Var _ x        -> Just $ Var (lookup' x env) x
    App _ t1 t2    -> do t1' <- go t1 
                         t2' <- go t2
                         app t1' t2'
    Lam _ x t b    -> do t' <- go t
                         lam x t' =<< goWith x (classifiedBy $ sort t') b
    Pi _ x t b     -> do t' <- go t                    
                         pi x t'  =<< goWith x (classifiedBy $ sort t') b
    Decl _ x t d b -> do t' <- go t
                         d' <- go d
                         b' <- goWith x (sort d') b
                         return $ Decl (sort b') x t' d' b'
    Quote _ t      -> do t' <- go t                     
                         return $ Quote (sort t') t'
    
{--
eraseSort :: Tm -> C.Tm
eraseSort e = 
  let go = eraseSort in
  case e of
    Var _ x        -> C.Var x
    App _ f a      -> C.App (go f) (go a)
    Lam _ x t b    -> C.Lam x (go t) (go b)
    Pi _ x t b     -> C.Pi x (go t) (go b)
    Decl _ x t d b -> C.Decl x (go t) (go d) (go b)
    Quote _ t      -> C.Quote (go t)
    Star           -> C.Star
    Box            -> C.Box
    Delta          -> C.Delta
--}