module Language.SystemU.Test.Core where

import Language.SystemU.Core as C

import Test.Framework (testGroup)
import Test.Framework.Providers.HUnit
import Test.HUnit

import Control.Applicative

isRight = either fail (const $ return True)

tests = testGroup "Core" [
    typecheckTests
  , substTests  
  , evalTests  
  ]
  
-- typecheck
typecheckTests = testGroup "typecheck" [
    testStar
  , testBox  
  , testPi  
  , testLam  
  , testTcFile "twice"
  , testTcFile "polyappStar"  
  , testTcFile "polyappBox"  
  , testTcFile "polyappDelta"      
  , testTcFile "U"  
  , testTcFile "Var"  
  , testTcFile "ProdStar"    
  , testTcFile "ProdBox"    
  , testTcFile "ProdDelta"        
  , testTcFile "Rep"  
  , testTcFile "AbsStar"  
  , testTcFile "AppStar"      
  , testTcFile "AbsBox"          
  , testTcFile "AppBox"          
  , testTcFile "AbsDelta"          
  , testTcFile "AppDelta"          
  , testTcFile "UId"  
--  , testTcFile "testUId"  
  , testTcFile "Exp"  
  , testTcFile "PExp"
  , testTcFile "mkRep"  
  , testTcFile "mkAbsStar"      
  , testTcFile "mkAppStar"          
  , testTcFile "mkAbsBox"      
  , testTcFile "mkAppBox"          
  , testTcFile "mkAbsDelta"      
  , testTcFile "mkAppDelta"          
    
{--  
  , testTcFile "testQuote"
  , testTcFile "testQuote2"    
  , testTcFile "testQuote3"        
  , testTcFile "testQuote4"            
  , testTcFile "testQuote5"                
--}
    
  , testTcFile "testQuoteTy"    
  , testTcFile "testQuoteTy2"    
  , testTcFile "testQuoteTy3"    
  , testTcFile "testQuoteTy4"        
  , testTcFile "testQuoteTy5"            
  , testTcFile "testQuoteTy6"                
    
  , testGroup "programs" [
      testTcPgmFile "U"
    , testTcPgmFile "Rep"
    , testTcPgmFile "mkAbsStar"      
    , testTcPgmFile "mkAppStar"          
    , testTcPgmFile "mkAbsBox"      
    , testTcPgmFile "mkAppBox"          
    , testTcPgmFile "mkAbsDelta"      
    , testTcPgmFile "mkAppDelta"          
    ]  
  ]

testStar = testCase "star" $ typecheck Star @?= Right Box
testBox  = testCase "box"  $ typecheck Box  @?= Right Delta

testPi   = testGroup "pi" [
    testCase "* -> *" $ typecheck (Pi "x" Star Star) @?= Right Box
  , testCase "# -> *" $ typecheck (Pi "x" Box  Star) @?= Right Box  
  , testCase "Pi x:*. x -> x" $ 
      typecheck (Pi "x" Star $ Pi "y" (Var "x") (Var "x")) @?= Right Star

  ]

testLam  = testGroup "lam" [
    testCase "\\x:*.x" $ typecheck (Lam "x" Star $ Var "x") @?= Right (Pi "x" Star Star)
  ]
           
testTcFile f = testCase f $ do
  e <- read <$> readFile ("examples/" ++ f)
  b <- isRight (typecheck e) 
  b @?= True
           
testTcPgmFile f = testCase f $ do
  e <- read <$> readFile ("examples/pgms/" ++ f)
  b <- isRight (typecheckPgm e) 
  b @?= True

-- subst
substTests = testGroup "subst" [
    testCase "free"  $ substitute "x" Star (Var "x") @?= Star
  , testSubstBound
  ]

testSubstBound = testCase "bound" $ 
  substitute "x" Star (Lam "x" Box (Var "x")) @?= (Lam "x" Box (Var "x"))
  
-- eval
evalTests = testGroup "eval" []
{--
    testCase "*" $ evaluate Star @?= Star
  , testCase "#" $ evaluate Box  @?= Box
  , testCase "^" $ evaluate Delta @?= Delta  
  , testCase "beta" $ evaluate beta @?= Star  
  ]
--}

beta = App (Lam "x" Box $ Var "x") Star

